/*!
 * \file rms.h
 * \brief C++ client library for the RMS API
 *
 * The rms_client allows easy communication to the RMS API.
 *
 * \author Russell Toris, WPI - rctoris@wpi.edu
 * \date December 30, 2012
 */

#ifndef RMS_H_
#define RMS_H_

#include <curl/curl.h>
#include <json/json.h>
#include <rmscpp/config/javascript_files/javascript_file.h>
#include <rmscpp/content/articles/article.h>
#include <rmscpp/content/content_pages/content_page.h>
#include <rmscpp/content/slides/slide.h>
#include <rmscpp/user_studies/experiments/experiment.h>
#include <rmscpp/user_studies/conditions/condition.h>
#include <rmscpp/user_studies/studies/study.h>
#include <rmscpp/user_studies/study_logs/study_log.h>
#include <rmscpp/users/user_accounts/user_account.h>
#include <string>
#include <vector>

/*!
 * \namespace rms
 * rms namespace for all RMS calls and objects
 */
namespace rms
{
/*!
 * \enum PROTOCOL
 * The different communication protocols that can be used with the RMS API.
 */
enum PROTOCOL
{
  HTTP_PROTOCOL, HTTPS_PROTOCOL
};

const std::string API = "/api/"; /*!< default script in the RMS API */

const std::string JAVASCRIPT_FILES_DEFAULT = "config/javascript_files/"; /*!< default Javascript files script in the RMS API */
const std::string JAVASCRIPT_FILES_ID_FIELD = "id"; /*!< Javascript file id field in the Javascript files script in the RMS API */

const std::string ARTICLES_DEFAULT = "content/articles/"; /*!< default articles script in the RMS API */
const std::string ARTICLES_ID_FIELD = "id"; /*!< article id field in the articles script in the RMS API */
const std::string ARTICLES_PAGES_ID_FIELD = "pageid"; /*!< page id field in the articles script in the RMS API */

const std::string CONTENT_PAGES_DEFAULT = "content/content_pages/"; /*!< default content pages script in the RMS API */
const std::string CONTENT_PAGES_ID_FIELD = "id"; /*!< content pages id field in the content pages script in the RMS API */

const std::string SLIDES_DEFAULT = "content/slides/"; /*!< default slides script in the RMS API */

const std::string CONDITIONS_DEFAULT = "user_studies/conditions/"; /*!< default conditions script in the RMS API */
const std::string CONDITIONS_ID_FIELD = "id"; /*!< condition id field in the conditions script in the RMS API */

const std::string EXPERIMENTS_DEFAULT = "user_studies/experiments/"; /*!< default experiments script in the RMS API */
const std::string EXPERIMENTS_ID_FIELD = "id"; /*!< experiment id field in the experiments script in the RMS API */

const std::string STUDIES_DEFAULT = "user_studies/studies/"; /*!< default studies script in the RMS API */
const std::string STUDIES_ID_FIELD = "id"; /*!< study id field in the studies script in the RMS API */

const std::string STUDY_LOGS_DEFAULT = "user_studies/study_logs/"; /*!< default study logs script in the RMS API */
const std::string STUDY_LOGS_ID_FIELD = "id"; /*!< study log id field in the study logs script in the RMS API */
const std::string STUDY_LOGS_EXPERIMENTS_ID_FIELD = "expid"; /*!< experiment id field in the study logs script in the RMS API */

const std::string USER_ACCOUNTS_DEFAULT = "users/user_accounts/"; /*!< default user_accounts script in the RMS API */
const std::string USER_ACCOUNTS_ID_FIELD = "id"; /*!< user account id field in the user_accounts script in the RMS API */

/*!
 * \struct rms_response
 * Buffer used when receiving data from the RMS API.
 */
typedef struct
{
  char *data; /*!< the data returned from the server */
  size_t size; /*!< the size of the data */
} rms_response;

/*!
 * \class rms_client
 * \brief Provides direct communication to the RMS API to via cURL.
 *
 * The rms_client handles communication to the RMS HTTP or HTTPS server using cURL.
 */
class rms_client
{
public:
  /*!
   * Creates a new client with the given information.
   *
   * @param host the host name of the server
   * @param port the port of the server (defaults to 80)
   * @param protocol the protocol used to communicate with the server (defaults to HTTP)
   */
  rms_client(std::string host, int port = 80, rms::PROTOCOL protocol = HTTP_PROTOCOL);
  /*!
   * Cleans up all resources left by cURL.
   */
  virtual ~rms_client();
  /*!
   * Sets the authentication header in cURL.
   *
   * @param username the username of the user
   * @param password the plain-text password for the user
   */
  void set_authorization(std::string username, std::string password);

  /*!
   * Get a vector of all downloaded Javascript files on the server.
   *
   * @return a vector of all downloaded Javascript files on the server
   */
  std::vector<rms::javascript_file> get_javascript_files();
  /*!
   * Get the Javascript file with the given ID number.
   *
   * @param fileid the Javascript file ID number
   * @return the Javascript file with the given ID number
   */
  rms::javascript_file get_javascript_file_by_id(int fileid);

  /*!
   * Get a vector of all articles on the server.
   *
   * @return a vector of all articles on the server
   */
  std::vector<rms::article> get_articles();
  /*!
   * Get the article with the given ID number.
   *
   * @param artid the article ID number
   * @return the article with the given ID number
   */
  rms::article get_article_by_id(int artid);
  /*!
   * Get a vector of all articles on the server that are assigned to the given page.
   *
   * @param pageid the page ID number
   * @return a vector of all articles on the server.
   */
  std::vector<rms::article> get_articles_by_pageid(int pageid);

  /*!
   * Get a vector of all content pages on the server.
   *
   * @return a vector of all content pages on the server
   */
  std::vector<rms::content_page> get_content_pages();
  /*!
   * Get the content page with the given ID number.
   *
   * @param pageid the content page ID number
   * @return the content page with the given ID number
   */
  rms::content_page get_content_page_by_id(int pageid);

  /*!
   * Get a vector of all slides on the server.
   *
   * @return a vector of all slides on the server
   */
  std::vector<rms::slide> get_slides();

  /*!
   * Get a vector of all conditions on the server.
   *
   * @return a vector of all conditions on the server
   */
  std::vector<rms::condition> get_conditions();
  /*!
   * Get the condition with the given ID number.
   *
   * @param condid the condition ID number
   * @return the condition with the given ID number
   */
  rms::condition get_condition_by_id(int condid);

  /*!
   * Get a vector of all experiments on the server.
   *
   * @return a vector of all experiments on the server
   */
  std::vector<rms::experiment> get_experiments();
  /*!
   * Get the experiment with the given ID number.
   *
   * @param expid the experiment ID number
   * @return the experiment with the given ID number
   */
  rms::experiment get_experiment_by_id(int expid);

  /*!
   * Get a vector of all studies on the server.
   *
   * @return a vector of all studies on the server
   */
  std::vector<rms::study> get_studies();
  /*!
   * Get the study with the given ID number.
   *
   * @param studyid the study ID number
   * @return the study with the given ID number
   */
  rms::study get_study_by_id(int studyid);

  /*!
   * Get a vector of all study log entries on the server.
   *
   * @return a vector of all study log entries on the server
   */
  std::vector<rms::study_log> get_study_logs();
  /*!
   * Get the study log entry with the given ID number.
   *
   * @param logid the study log entry ID number
   * @return the study log entry with the given ID number
   */
  rms::study_log get_study_log_by_id(int logid);
  /*!
   * Get a vector of all study log entries with the given experiment ID number.
   *
   * @param logid the experiment ID number
   * @return a vector of all study log entries with the given experiment ID number
   */
  std::vector<rms::study_log> get_study_logs_by_expid(int expid);

  /*!
   * Get a vector of all user accounts on the server.
   *
   * @return a vector of all user accounts on the server
   */
  std::vector<rms::user_account> get_user_accounts();
  /*!
   * Get the user account with the given ID number.
   *
   * @param userid the user account ID number
   * @return the user account with the given ID number
   */
  rms::user_account get_user_account_by_id(int userid);
private:
  std::string url_start; /*!< the base string for the URL */
  CURL *curl; /*!< used to communicate with the HTTP or HTTPS server */

  /*!
   * Parse the given JSON object as an array of Javascript files.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of Javascript files that was parsed from the JSON object
   */
  std::vector<rms::javascript_file> parse_javascript_files_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a Javascript file.
   *
   * @param json a pointer to the JSON object
   * @return the Javascript file that was parsed from the JSON object
   */
  rms::javascript_file parse_javascript_file(json_object *json);

  /*!
   * Parse the given JSON object as an array of articles.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of articles that was parsed from the JSON object
   */
  std::vector<rms::article> parse_articles_array(json_object *json_array);
  /*!
   * Parse the given JSON object as an article.
   *
   * @param json a pointer to the JSON object
   * @return the article that was parsed from the JSON object
   */
  rms::article parse_article(json_object *json);

  /*!
   * Parse the given JSON object as an array of content pages.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of content pages that was parsed from the JSON object
   */
  std::vector<rms::content_page> parse_content_pages_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a content page.
   *
   * @param json a pointer to the JSON object
   * @return the content page that was parsed from the JSON object
   */
  rms::content_page parse_content_page(json_object *json);

  /*!
   * Parse the given JSON object as an array of slides.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of slides that was parsed from the JSON object
   */
  std::vector<rms::slide> parse_slides_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a slide.
   *
   * @param json a pointer to the JSON object
   * @return the slide that was parsed from the JSON object
   */
  rms::slide parse_slide(json_object *json);

  /*!
   * Parse the given JSON object as an array of conditions.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of conditions that was parsed from the JSON object
   */
  std::vector<rms::condition> parse_conditions_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a condition.
   *
   * @param json a pointer to the JSON object
   * @return the condition that was parsed from the JSON object
   */
  rms::condition parse_condition(json_object *json);

  /*!
   * Parse the given JSON object as an array of experiments.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of experiments that was parsed from the JSON object
   */
  std::vector<rms::experiment> parse_experiments_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a condition.
   *
   * @param json a pointer to the JSON object
   * @return the experiment that was parsed from the JSON object
   */
  rms::experiment parse_experiment(json_object *json);

  /*!
   * Parse the given JSON object as an array of studies.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of studies that was parsed from the JSON object
   */
  std::vector<rms::study> parse_studies_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a study.
   *
   * @param json a pointer to the JSON object
   * @return the study that was parsed from the JSON object
   */
  rms::study parse_study(json_object *json);

  /*!
   * Parse the given JSON object as an array of study log entries.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of study log entries that was parsed from the JSON object
   */
  std::vector<rms::study_log> parse_study_logs_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a study log entry.
   *
   * @param json a pointer to the JSON object
   * @return the study log entry that was parsed from the JSON object
   */
  rms::study_log parse_study_log(json_object *json);

  /*!
   * Parse the given JSON object as an array of user accounts.
   *
   * @param json_array a pointer to the JSON object
   * @return the vector of user accounts that was parsed from the JSON object
   */
  std::vector<rms::user_account> parse_user_accounts_array(json_object *json_array);
  /*!
   * Parse the given JSON object as a user account.
   *
   * @param json a pointer to the JSON object
   * @return the user account that was parsed from the JSON object
   */
  rms::user_account parse_user_account(json_object *json);

  /*!
   * Extract a string from the given JSON object with the given key.
   *
   * @param json a pointer to the JSON object
   * @param key the key of the string in the JSON object
   * @return the string that was extracted or "NULL" if it was a null value
   */
  std::string extract_json_string(json_object *json, std::string key);
  /*!
   * Extract a boolean from the given JSON object with the given key.
   *
   * @param json a pointer to the JSON object
   * @param key the key of the boolean in the JSON object
   * @return the boolean that was extracted
   */
  bool extract_json_bool(json_object *json, std::string key);
  /*!
   * Extract an int from the given JSON object with the given key.
   *
   * @param json a pointer to the JSON object
   * @param key the key of the int in the JSON object
   * @return the int that was extracted
   */
  int extract_json_int(json_object *json, std::string key);

  /*!
   * Make a request to the RMS server with the given URL.
   *
   * @param url the URL to query
   * @return the JSON object that was returned from the server
   */
  json_object *make_request(std::string url);
  /*!
   * Frees all memory used by the given rms_response.
   * @param buf the buffer to free
   */
  void free_rms_response(rms::rms_response *buf);
};

/*!
 * The callback function used by cURL to store the response from the Rovio.
 * This function should only be used by cURL internally by the rms_client.
 *
 * @param ptr the pointer to the data
 * @param size size * nmemb is the exact size of the data
 * @param nmemb size * nmemb is the exact size of the data
 * @param buf the buffer to put the data into
 */
size_t write_data(char *ptr, size_t size, size_t nmemb, rms::rms_response *buf);
}

#endif
